@@ -2,9 +2,8 @@
  *
  *  OBEX Server
  *
- *  Copyright (C) 2007-2008  Nokia Corporation
- *  Copyright (C) 2007-2008  Instituto Nokia de Tecnologia (INdT)
- *  Copyright (C) 2007-2009  Marcel Holtmann <marcel@holtmann.org>
+ *  Copyright (C) 2007-2010  Nokia Corporation
+ *  Copyright (C) 2007-2010  Marcel Holtmann <marcel@holtmann.org>
  *
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -68,8 +67,10 @@ typedef struct {
 static void os_reset_session(struct obex_session *os)
 {
 	if (os->object) {
+		os->driver->set_io_watch(os->object, NULL, NULL);
 		os->driver->close(os->object);
 		os->object = NULL;
+		os->obj = NULL;
 		if (os->aborted && os->cmd == OBEX_CMD_PUT && os->current_folder) {
 			gchar *path;
 			path = g_build_filename(os->current_folder, os->name, NULL);
@@ -90,6 +91,7 @@ static void os_reset_session(struct obex_session *os)
 		g_free(os->buf);
 		os->buf = NULL;
 	}
+
 	os->driver = NULL;
 	os->aborted = FALSE;
 	os->offset = 0;
@@ -386,10 +388,9 @@ int os_prepare_get(struct obex_session *os, gchar *filename, size_t *size)
 	gint err;
 	gpointer object;
 
-	object = os->driver->open(filename, O_RDONLY, 0, size);
+	object = os->driver->open(filename, O_RDONLY, 0, size, &err);
 	if (object == NULL) {
-		err = -errno;
-		error("open(%s): %s (%d)", filename, strerror(errno), errno);
+		error("open(%s): %s (%d)", filename, strerror(-err), -err);
 		goto fail;
 	}
 
@@ -432,11 +433,13 @@ static gint obex_write_stream(struct obex_session *os,
 
 	len = os->driver->read(os->object, os->buf, os->tx_mtu);
 	if (len < 0) {
-		gint err = errno;
-		error("read(): %s (%d)", strerror(err), err);
+		error("read(): %s (%d)", strerror(-len), -len);
+		if (len == -EAGAIN)
+			return len;
+
 		g_free(os->buf);
 		os->buf = NULL;
-		return -err;
+		return len;
 	}
 
 	ptr = os->buf;
@@ -465,13 +468,14 @@ gint os_prepare_put(struct obex_session *os)
 {
 	gchar *path;
 	gint len;
+	int err;
 
 	path = g_build_filename(os->current_folder, os->name, NULL);
 	os->object = os->driver->open(path, O_WRONLY | O_CREAT | O_TRUNC, 0600,
 					os->size != OBJECT_SIZE_UNKNOWN ?
-					(size_t *) &os->size : NULL);
+					(size_t *) &os->size : NULL, &err);
 	if (os->object == NULL) {
-		error("open(%s): %s (%d)", path, strerror(errno), errno);
+		error("open(%s): %s (%d)", path, strerror(-err), -err);
 		g_free(path);
 		return -EPERM;
 	}
@@ -489,13 +493,11 @@ gint os_prepare_put(struct obex_session *os)
 
 		w = os->driver->write(os->object, os->buf + len, os->offset - len);
 		if (w < 0) {
-			gint err = errno;
-			error("write(%s): %s (%d)", path, strerror(errno),
-									errno);
-			if (err == EINTR)
+			error("write(%s): %s (%d)", path, strerror(-w), -w);
+			if (w == -EINTR)
 				continue;
 			else
-				return -err;
+				return w;
 		}
 
 		len += w;
@@ -544,11 +546,10 @@ static gint obex_read_stream(struct obex_session *os, obex_t *obex,
 
 		w = os->driver->write(os->object, buffer + len, size - len);
 		if (w < 0) {
-			gint err = errno;
-			if (err == EINTR)
+			if (w == -EINTR)
 				continue;
 			else
-				return -err;
+				return w;
 		}
 
 		len += w;
@@ -559,6 +560,44 @@ static gint obex_read_stream(struct obex_session *os, obex_t *obex,
 	return 0;
 }
 
+static gboolean handle_async_io(gpointer object, int flags, int err,
+				gpointer user_data)
+{
+	struct obex_session *os = user_data;
+	int ret = 0;
+
+	if (err < 0) {
+		ret = err;
+		goto proceed;
+	}
+
+	if (flags & (G_IO_IN | G_IO_PRI))
+		ret = obex_write_stream(os, os->obex, os->obj);
+	else if (flags & G_IO_OUT)
+		ret = obex_read_stream(os, os->obex, os->obj);
+
+proceed:
+	switch (ret) {
+	case -EINVAL:
+		OBEX_ObjectSetRsp(os->obj, OBEX_RSP_BAD_REQUEST,
+				OBEX_RSP_BAD_REQUEST);
+	case -EPERM:
+		OBEX_ObjectSetRsp(os->obj, OBEX_RSP_FORBIDDEN,
+					OBEX_RSP_FORBIDDEN);
+		break;
+	default:
+		if (ret < 0)
+			OBEX_ObjectSetRsp(os->obj,
+					OBEX_RSP_INTERNAL_SERVER_ERROR,
+					OBEX_RSP_INTERNAL_SERVER_ERROR);
+		break;
+	}
+
+	OBEX_ResumeRequest(os->obex);
+
+	return FALSE;
+}
+
 static gboolean check_put(obex_t *obex, obex_object_t *obj)
 {
 	struct obex_session *os;
@@ -645,7 +684,7 @@ static gboolean check_put(obex_t *obex, obex_object_t *obj)
 		}
 	}
 
-	if (!os->service || !os->service->chkput)
+	if (!os->service->chkput)
 		goto done;
 
 	ret = os->service->chkput(obex, obj);
@@ -659,6 +698,11 @@ static gboolean check_put(obex_t *obex, obex_object_t *obj)
 	case -EPERM:
 		OBEX_ObjectSetRsp(obj, OBEX_RSP_FORBIDDEN, OBEX_RSP_FORBIDDEN);
 		return FALSE;
+	case -EAGAIN:
+		OBEX_SuspendRequest(obex, obj);
+		os->obj = obj;
+		os->driver->set_io_watch(os->object, handle_async_io, os);
+		return TRUE;
 	default:
 		debug("Unhandled chkput error: %d", ret);
 		OBEX_ObjectSetRsp(obj, OBEX_RSP_INTERNAL_SERVER_ERROR,
@@ -727,13 +771,12 @@ static void obex_event(obex_t *obex, obex_object_t *obj, gint mode,
 		case OBEX_CMD_PUT:
 		case OBEX_CMD_GET:
 		case OBEX_CMD_SETPATH:
+		default:
 			os_session_mark_aborted(os);
 			if (os->service->reset)
 				os->service->reset(obex);
 			os_reset_session(os);
 			break;
-		default:
-			break;
 		}
 		break;
 	case OBEX_EV_REQHINT:
@@ -758,7 +801,7 @@ static void obex_event(obex_t *obex, obex_object_t *obj, gint mode,
 	case OBEX_EV_REQCHECK:
 		switch (cmd) {
 		case OBEX_CMD_PUT:
-			if (os->service && os->service->put)
+			if (os->service)
 				check_put(obex, obj);
 			break;
 		default:
@@ -796,6 +839,11 @@ static void obex_event(obex_t *obex, obex_object_t *obj, gint mode,
 			OBEX_ObjectSetRsp(obj,
 				OBEX_RSP_FORBIDDEN, OBEX_RSP_FORBIDDEN);
 			break;
+		case -EAGAIN:
+			OBEX_SuspendRequest(obex, obj);
+			os->obj = obj;
+			os->driver->set_io_watch(os->object, handle_async_io, os);
+			break;
 		default:
 			OBEX_ObjectSetRsp(obj,
 				OBEX_RSP_INTERNAL_SERVER_ERROR,
@@ -805,7 +853,21 @@ static void obex_event(obex_t *obex, obex_object_t *obj, gint mode,
 
 		break;
 	case OBEX_EV_STREAMEMPTY:
-		obex_write_stream(os, obex, obj);
+		switch (obex_write_stream(os, obex, obj)) {
+		case -EPERM:
+			OBEX_ObjectSetRsp(obj,
+				OBEX_RSP_FORBIDDEN, OBEX_RSP_FORBIDDEN);
+			break;
+		case -EAGAIN:
+			OBEX_SuspendRequest(obex, obj);
+			os->obj = obj;
+			os->driver->set_io_watch(os->object, handle_async_io,
+							os);
+			break;
+		default:
+			break;
+		}
+
 		break;
 	case OBEX_EV_LINKERR:
 		break;
